package it.cavallium.filequeue;

import java.time.Duration;
import java.util.concurrent.atomic.AtomicLong;
import java.util.concurrent.locks.LockSupport;

class QueueToConsumer<T> implements IQueueToConsumer<T> {

	private final long BACKOFF_NS = Duration.ofMillis(2).toNanos();
	private final long MAX_BACKOFF_NS = Duration.ofMillis(500).toNanos();

	private final long HALF_SECOND_NS = Duration.ofMillis(500).toNanos();
	private final AtomicLong preAddQueued;
	private final AtomicLong afterAddQueued;
	private final SimpleQueue<T> queue;
	private final QueueConsumer<T> consumer;
	private Manager manager;
	private volatile boolean closed;
	private volatile Throwable managerFailure = null;

	public QueueToConsumer(SimpleQueue<T> queue, QueueConsumer<T> consumer) {
		this.queue = queue;
		this.consumer = consumer;
		this.preAddQueued = new AtomicLong(queue.size());
		this.afterAddQueued = new AtomicLong(queue.size());
	}

	protected String getQueueManagerName() {
		return "queue-manager";
	}

	public synchronized void startQueue() {
		if (managerFailure != null) {
			manager = null;
			managerFailure = null;
		}
		if (manager == null) {
			this.manager = new Manager();
			manager.setName(getQueueManagerName());
			manager.start();
		}
	}

	@Override
	public void add(T value) {
		checkManagerStatus();
		boolean shouldAdd = true;
		if (preAddQueued.getAndIncrement() == 0) {
			boolean crashed = true;
			try {
				if (consumer.tryConsume(value)) {
					shouldAdd = false;
					preAddQueued.decrementAndGet();
				}
				crashed = false;
			} finally {
				if (crashed) {
					shouldAdd = false;
					preAddQueued.decrementAndGet();
				}
			}
		}
		if (shouldAdd && !closed) {
			queue.add(value);
			var queueSize = afterAddQueued.incrementAndGet();
			if (queueSize == 1) {
				LockSupport.unpark(manager);
			}
		}
	}

	private void checkManagerStatus() {
		if (managerFailure != null) {
			throw new RuntimeException("Can't add an element, the queue manager has failed", managerFailure);
		}
	}

	@Override
	public void close() {
		closed = true;
	}

	private class Manager extends Thread {

		@Override
		public void run() {
			try {
				while (!closed) {
					runStepInternal(true);
				}
			} catch (Throwable ex) {
				managerFailure = ex;
				throw ex;
			} finally {
				closed = true;
			}
		}
	}

	public void runStep(boolean park) {
		if (manager != null) throw new UnsupportedOperationException("This queue is managed");
		runStepInternal(false);
	}

	private void runStepInternal(boolean park) {
		T element;
		boolean shouldRemove = afterAddQueued.getAndUpdate(n -> n > 0 ? n - 1 : 0) > 0;
		if (shouldRemove) {
			element = queue.remove();
			long nextDelay = BACKOFF_NS;
			while (!closed && !consumer.tryConsume(element)) {
				if (park) {
					LockSupport.parkNanos(nextDelay);
				}
				if (nextDelay + BACKOFF_NS <= MAX_BACKOFF_NS) {
					nextDelay += BACKOFF_NS;
				} else if (nextDelay < MAX_BACKOFF_NS) {
					nextDelay = MAX_BACKOFF_NS;
				}
			}
			preAddQueued.updateAndGet(n -> n > 0 ? n - 1 : 0);
		} else if (park) {
			LockSupport.parkNanos(HALF_SECOND_NS);
		}
	}
}
