package it.cavallium.filequeue;

import java.io.Closeable;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.MappedByteBuffer;
import java.nio.channels.FileChannel;
import java.nio.channels.FileChannel.MapMode;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.StandardOpenOption;
import java.util.NoSuchElementException;
import java.util.concurrent.atomic.AtomicInteger;

public class SimpleQueueFileChannelFixedSize implements Closeable {

	private final Path filePath;
	private final FileChannel ch;
	private final MappedByteBuffer writer;
	private final long fixedSize;
	private int readPosition = 0;
	private int writePosition = 0;
	private final AtomicInteger size = new AtomicInteger();

	public SimpleQueueFileChannelFixedSize(long fixedSize, Path filePath) throws IOException {
		this.filePath = filePath;
		this.fixedSize = fixedSize;
		Files.deleteIfExists(filePath);
		this.ch = FileChannel.open(filePath, StandardOpenOption.CREATE_NEW, StandardOpenOption.WRITE, StandardOpenOption.READ);
		ch.truncate(fixedSize);
		this.writer = ch.map(MapMode.READ_WRITE, 0L, fixedSize);
	}

	@Override
	public void close() throws IOException {
		ch.close();
		Files.deleteIfExists(filePath);
	}

	public void add(ByteBuffer element) {
		if (!offer(element)) {
			throw new IllegalStateException("Can't add the element, not enough space");
		}
	}

	/**
	 * @return true if there is enough space, false otherwise
	 */
	public boolean offer(ByteBuffer element) {
		int size = Math.toIntExact(element.remaining());
		if (writePosition + size + Integer.BYTES > fixedSize) {
			if (writePosition == 0) throw new UnsupportedOperationException("Entity too large! " + size + " bytes");
			return false;
		}
		writer.slice(writePosition, Integer.BYTES).putInt(size);
		writer.put(writePosition + Integer.BYTES, element, 0, size);
		writePosition += size + Integer.BYTES;
		this.size.incrementAndGet();
		return true;
	}

	public ByteBuffer poll() {
		if (readPosition >= fixedSize) {
			return null;
		}
		if (size.decrementAndGet() < 0) {
			size.incrementAndGet();
			return null;
		}
		int size = writer.slice(readPosition, Integer.BYTES).getInt();
		var segment = ByteBuffer.allocate(size);
		segment.put(0, writer, readPosition + Integer.BYTES, size);
		readPosition += size + Integer.BYTES;
		return segment;
	}

	public ByteBuffer remove() {
		var polled = poll();
		if (polled == null) throw new NoSuchElementException();
		return polled;
	}

	public int size() {
		return Math.max(0, size.get());
	}

	public void hibernate() {
		try {
			this.ch.force(false);
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
}
