package it.cavallium.filequeue;

import java.io.Closeable;
import java.io.IOException;
import java.lang.foreign.Arena;
import java.lang.foreign.MemorySegment;
import java.lang.foreign.ValueLayout;
import java.nio.channels.FileChannel;
import java.nio.channels.FileChannel.MapMode;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.StandardOpenOption;
import java.util.NoSuchElementException;
import java.util.concurrent.atomic.AtomicInteger;

public class SimpleQueueMemorySegmentFixedSize implements Closeable {

	private final Path filePath;
	private final FileChannel ch;
	private final MemorySegment writer;
	private final Arena mmapArena;
	private final long fixedSize;
	private long readPosition = 0;
	private long writePosition = 0;
	private final AtomicInteger size = new AtomicInteger();

	public SimpleQueueMemorySegmentFixedSize(long fixedSize, Path filePath) throws IOException {
		this.mmapArena = Arena.ofShared();
		this.filePath = filePath;
		this.fixedSize = fixedSize;
		Files.deleteIfExists(filePath);
		this.ch = FileChannel.open(filePath, StandardOpenOption.CREATE_NEW, StandardOpenOption.WRITE, StandardOpenOption.READ);
		ch.truncate(fixedSize);
		this.writer = ch.map(MapMode.READ_WRITE, 0L, fixedSize, mmapArena);
	}

	@Override
	public void close() throws IOException {
		try {
			ch.close();
			Files.deleteIfExists(filePath);
		} finally {
			mmapArena.close();
		}
	}

	public void add(MemorySegment element) {
		if (!offer(element)) {
			throw new IllegalStateException("Can't add the element, not enough space");
		}
	}

	/**
	 * @return true if there is enough space, false otherwise
	 */
	public boolean offer(MemorySegment element) {
		int size = Math.toIntExact(element.byteSize());
		if (writePosition + size + Integer.BYTES > fixedSize) {
			if (writePosition == 0) throw new UnsupportedOperationException("Entity too large! " + size + " bytes");
			return false;
		}
		writer.set(ValueLayout.JAVA_INT_UNALIGNED, writePosition, size);
		MemorySegment.copy(element, 0, writer, writePosition + Integer.BYTES, size);
		writePosition += size + Integer.BYTES;
		this.size.incrementAndGet();
		return true;
	}

	public MemorySegment poll(Arena arena) {
		if (readPosition >= fixedSize) {
			return null;
		}
		if (size.decrementAndGet() < 0) {
			size.incrementAndGet();
			return null;
		}
		int size = writer.get(ValueLayout.JAVA_INT_UNALIGNED, readPosition);
		var segment = arena.allocate(size);
		MemorySegment.copy(writer, readPosition + Integer.BYTES, segment, 0, size);
		readPosition += size + Integer.BYTES;
		return segment;
	}

	public MemorySegment remove(Arena arena) {
		var polled = poll(arena);
		if (polled == null) throw new NoSuchElementException();
		return polled;
	}

	public int size() {
		return Math.max(0, size.get());
	}

	public void hibernate() {
		this.writer.unload();
	}
}
